import es
import playerlib
import effectlib
import popuplib
import cfglib
import cmdlib
import urllib2
import gamethread
import usermsg
import vecmath
import ConfigParser
import cPickle
import time

#AddonInfo:
info = es.AddonInfo() 
info.name     	 = "Double Domination" 
info.basename 	 = "double_domination"
info.version  	 = "v1.0"
info.url      	 = "www.NextGenerationGaming.de" 
info.description = "Double Domination game mode from Unreal Tournament."
info.author   	 = "Ruuubi"

#Classes:
''' Creates addon settings and executes the config.cfg. '''
class ConfigManager(object):
	def __init__(self):
		es.ServerVar('dd_admins','','')
		self.minPlayers = es.ServerVar('dd_min_players',2,'')
		self.dominatingTime = es.ServerVar('dd_dominating_time',10,'')
		self.mode = es.ServerVar('dd_mode',3,'')
		self.timeLimit = es.ServerVar('dd_time_limit',10,'')
		self.pointLimit = es.ServerVar('dd_point_limit',5,'')
		self.showHud = es.ServerVar('dd_hud',1,'')

		self.dmEnabled = es.ServerVar('dd_dm_enabled',1,'')
		es.set('dd_dm_mode',1,'')
		es.set('dd_dm_delay',5,'')

		self.spEnabled = es.ServerVar('dd_sp_time',3,'')
		es.set('dd_sp_color_enabled',1,'')
		es.set('dd_sp_noblock',1,'')

		self.cashSpawn = es.ServerVar('dd_cash_spawn',500,'')
		self.cashCapture = es.ServerVar('dd_cash_team_capture',1000,'')
		self.cashDominate = es.ServerVar('dd_cash_team_dominate',1000,'')
		self.cashScore = es.ServerVar('dd_cash_team_score',2000,'')

		self.scoreCapture = es.ServerVar('dd_score_team_capture',0,'')
		self.scoreDominate = es.ServerVar('dd_score_team_dominate',0,'')
		self.scoreScore = es.ServerVar('dd_score_team_score',0,'')

		self.announcerEnabled = es.ServerVar('dd_announcer_enabled',1,'')
		es.ServerVar('dd_announcer_commands','','')
		self.announcerMain = es.ServerVar('dd_announcer_main','announcer_evil','')
		self.announcerSavetime = es.ServerVar('dd_announcer_savetime',30,'')

		self.weaponsRemove = es.ServerVar('dd_misc_weapons_remove',1,'')
		self.weaponsAutoclean = es.ServerVar('dd_misc_weapons_autoclean',30,'')
		self.drawBeams = es.ServerVar('dd_misc_draw_beams',1,'')

		cfglib.AddonCFG(PATH_ADDON+'/configs/config.cfg').execute()

		self.admins = str(es.ServerVar('dd_admins')).split(';')
		self.announcerCommands = str(es.ServerVar('dd_announcer_commands')).split(';')

''' Controls the A and B checkpoint. '''
class Checkpoint(object):
	''' Initialize the checkpoint '''
	def __init__(self,name):
		self.index = None										# Entityindex of checkpoint
		self.name = name										# Name of checkpoint (A or B)
		self.team = 0											# 0 = Neutral (Can not be taken) | 1 = Neutral | 2 = Red | 3 = Blue
		self.loc = None											# X,Y,Z location
		self.hitbox = {'start': None, 'end': None}				# Hitbox boundaries
		self.hitboxPlayerlist = []								# List of players currently touching the checkpoint
		self.lastTouched = 0									# Userid of last touching player
		self.colors = {0: {'col': '255 255 255', 'amt': 0},		# Color and alpha settings in state 0-3 to show the owner [invisible, white, red, blue]
						1: {'col': '155 155 155', 'amt': 255},
						2: {'col': '255 0 0', 'amt': 255},
						3: {'col': '0 80 255', 'amt': 255}}

	''' Spawn the checkpoint (prop_dynamic) '''
	def spawn(self):
		entList = es.createentitylist()
		for entIndex in dict(entList):
			if es.entitygetvalue(entIndex,'targetname') == ('__dd_checkpoint_%s' % self.name): es.server.queuecmd('es_xremove %s' % entIndex)
		self.team = 0
		es.precachemodel(PATH_CHECKPOINT_A) if self.name == 'A' else es.precachemodel(PATH_CHECKPOINT_B)
		index = es.createentity('prop_dynamic')
		es.entitysetvalue(index,'targetname','__dd_checkpoint_%s' % self.name)
		es.entitysetvalue(index,'model',PATH_CHECKPOINT_A) if self.name == 'A' else es.entitysetvalue(index,'model',PATH_CHECKPOINT_B)
		es.entitysetvalue(index,'origin','%s %s %s' % (self.loc[0],self.loc[1],self.loc[2]))
		es.entitysetvalue(index,'renderamt',self.colors[self.team]['amt'])
		es.entitysetvalue(index,'rendercolor',self.colors[self.team]['col'])
		es.entitysetvalue(index,'rendermode','4')
		es.entitysetvalue(index,'disableshadows',1)
		self.index = index
		es.server.queuecmd('es_xspawnentity %s' % index)

	''' Set X,Y,Z location and hitbox boundaries '''
	def setLoc(self,(x,y,z)):
		self.loc = [x,y,z]
		self.hitbox['start'] = [x-40,y-40,z]
		self.hitbox['end'] = [x+40,y+40,z+100]
		self.spawn()

	''' Set team and change color/alpha '''
	def setTeam(self,teamnumber):
		self.team = int(teamnumber)
		es.entitysetvalue(self.index,'renderamt',self.colors[self.team]['amt'])
		es.entitysetvalue(self.index,'rendercolor',self.colors[self.team]['col'])

	''' Draw a beam into the sky (if enabled) '''
	def drawBeam(self):
		if self.team == 0: return
		r,g,b = self.colors[self.team]['col'].split(' ')
		#effectlib.drawBox(self.hitbox['start'],self.hitbox['end'],red=0,green=255,blue=0,brightness=96,seconds=0.3)
		effectlib.drawLine(self.loc,('%s,%s,%s' % (self.loc[0],self.loc[1],self.loc[2]+4500)),red=r,green=g,blue=b,brightness=100,seconds=0.3)

	''' Check if somebody is touching the checkpoint. If yes: Add him to the hitboxPlayerlist and set lastTouched '''
	def checkHitboxStart(self,userid,loc):
		if self.team == 0: return
		if vecmath.isbetweenRect(loc,self.hitbox['start'],self.hitbox['end']):
			self.lastTouched = userid
			self.hitboxPlayerlist.append(es.getplayerteam(userid))

	''' Change the owner only if one team is touching it, not both teams at the same time '''
	def checkHitboxEnd(self):
		if self.team == 0 or self.hitboxPlayerlist[1:] != self.hitboxPlayerlist[:-1] or len(self.hitboxPlayerlist) < 1: return
		newTeam = self.hitboxPlayerlist[0]
		if newTeam != self.team:
			self.setTeam(newTeam)
			es.emitsound('entity',self.index,'ruuubi/dd/misc/alarm.mp3',0.2,0.2)
			if config.scoreCapture: es.server.queuecmd('score add %i %i'% (self.lastTouched,config.scoreCapture))
			if config.cashCapture: playerlib.getPlayer(self.lastTouched).cash += config.cashCapture
			game.changedOwner()

''' Controls delays and their execute function '''
class DelayManager(object):
	''' Initialize delaylist and prefix '''
	def __init__(self):
		self.prefix = 'DoubleDomination_'
		self.delays = []

	''' Take a delayname and return it with a prefix '''
	def addPrefix(self,name):
		return self.prefix + str(name)

	''' Add a delay with a name and duration to the list '''
	def add(self,name,duration):
		if name in self.delays: gamethread.cancelDelayed(self.addPrefix(name))
		else: self.delays.append(name)
		gamethread.delayedname(duration,self.addPrefix(name),self.durationEnd,name)

	''' Remove a delay with the name from the list and cancel it '''
	def remove(self,name):
		if name in self.delays:
			gamethread.cancelDelayed(self.addPrefix(name))
			self.delays.remove(name)

	''' Remove all delays from the list and cancel them '''
	def removeAll(self):
		for x in self.delays: gamethread.cancelDelayed(self.addPrefix(x))
		self.delays = []

	''' Remove all delays from the list and cancel them as long as they are not given as an argument '''
	def removeAllExcept(self,delayList):
		for delay in list(self.delays):
			if delay not in delayList: self.remove(delay)

	''' This function will handle all expired delays, remove them from the list and execute other functions '''
	def durationEnd(self,name):
		if name in self.delays:
			self.delays.remove(name)
			if name == 'tick':
				self.add('tick',0.1)
				game.tick()
			elif name == 'gametime':
				game.reduceTime()
			elif name == 'dominatingCountdown':
				if game.dominatingTime < 1:
					game.teamScores()
				else:
					game.centermsg('%s..' % game.dominatingTime)
					announcer.playSound('countdown_0%s' % game.dominatingTime)
					game.dominatingTime -= 1
					self.add('dominatingCountdown',1)
			elif name == 'newRound':
				game.centermsg('New round in...')
				announcer.playSound('new_round_in')
				self.add('newRoundCD',2)
			elif name == 'newRoundCD':
				if game.newRoundCD < 1:
					game.newRound()
				else:
					game.centermsg('%s..' % game.newRoundCD)
					announcer.playSound('countdown_0%s' % game.newRoundCD)
					game.newRoundCD -= 1
					self.add('newRoundCD',1)
			elif name == 'weaponCleanup':
				entDict = es.createentitylist()
				for entIndex in filter(lambda x: entDict[x]['classname'].startswith('weapon_'),entDict):
					if es.getindexprop(entIndex,'CBaseEntity.m_hOwnerEntity') == -1: es.server.queuecmd('es_xremove %s' % entIndex)
				if config.weaponsAutoclean: delay.add('weaponCleanup',config.weaponsAutoclean)
			elif name == 'flawlessVictory':
				announcer.playSound('flawless_victory')

''' Saves player in a dictionary to handle player settings (like steamid, announcer, text settings) '''
class PlayerManager(dict):
	''' Return data from player object '''
	def __getitem__(self,userid):
		userid = int(userid)
		if userid not in self: self[userid] = PlayerObject(userid)
		return super(PlayerManager,self).__getitem__(userid)

	''' Call "delitem" function on player object '''
	def __delitem__(self,userid):
		userid = int(userid)
		if userid in self: super(PlayerManager,self).__delitem__(userid)

''' Player object used by the PlayerManager '''
class PlayerObject(object):
	''' Initialize player object and add them to the database or take their previous settings from the database if they are already in it '''
	def __init__(self,userid):
		global database
		self.userid = int(userid)
		self.steamid = es.getplayersteamid(userid)
		if self.steamid in database:
			self.announcer = database[self.steamid]['announcer']
			self.text = database[self.steamid]['text']
		else:
			self.announcer = str(config.announcerMain)
			self.text = True
			database[self.steamid] = {}
			database[self.steamid]['announcer'] = self.announcer
			database[self.steamid]['text'] = self.text
		database[self.steamid]['lastcon'] = int(time.time())

''' Handles all other functions that belong to the game '''
class GameManager(object):
	''' Initialize the game '''
	def __init__(self):
		self.enabled = False 		# Is the map playable [Valid Checkpoint locations]
		self.roundStarted = False	# Is the game/round running
		self.reset()
		self.screenoverlays = {'1|1': 'overlays/ruuubi/dd/info_white_white',
								'1|2': 'overlays/ruuubi/dd/info_white_red',
								'1|3': 'overlays/ruuubi/dd/info_white_blue',
								'2|1': 'overlays/ruuubi/dd/info_red_white',
								'2|2': 'overlays/ruuubi/dd/info_red_red',
								'2|3': 'overlays/ruuubi/dd/info_red_blue',
								'3|1': 'overlays/ruuubi/dd/info_blue_white',
								'3|2': 'overlays/ruuubi/dd/info_blue_red',
								'3|3': 'overlays/ruuubi/dd/info_blue_blue'}

	''' Reset the game '''
	def reset(self):
		self.dominatingTeam = 0								# Team that is currently dominating: 2 = Red | 3 = Blue
		self.dominatingTeamText = ''						# Will either be 'Red' or 'Blue'
		self.dominatingTime = config.dominatingTime-2		# Delay until the dominating team scores
		self.newRoundCD = 5									# Delay until a new round starts

	''' Open mapdata.ini to check for valid locations '''
	def loadMapData(self):
		self.enabled = False
		self.roundStarted = False
		self.reset()
		curMap = str(es.ServerVar('eventscripts_currentmap'))
		data = ConfigParser.ConfigParser()
		data.optionxform = str
		data.read(PATH_ADDON+'/database/mapdata.ini')
		for mapname in data.sections():
			if mapname == curMap:
				if 'checkpoint_a' in data.options(mapname): pointA.setLoc(map(int,data.get(mapname,'checkpoint_a').split(',')))
				if 'checkpoint_b' in data.options(mapname): pointB.setLoc(map(int,data.get(mapname,'checkpoint_b').split(',')))
		if pointA.loc != None and pointB.loc != None: self.enabled = True

	''' Set downloads for players (sounds/models/materials) '''
	def setDownloads(self):
		for file in downloadList: es.stringtable('downloadables',file)

	''' Start a new round if the map is playable '''
	def startRound(self):
		self.loadMapData()
		if self.enabled:
			if es.getplayercount(2)+es.getplayercount(3) >= config.minPlayers:
				if config.weaponsRemove: delay.add('weaponCleanup',0.1)
				elif config.weaponsAutoclean: delay.add('weaponCleanup',config.weaponsAutoclean)
				es.server.queuecmd('mp_ignore_round_win_conditions 1')
				entDict = es.createentitylist()
				for entIndex in entDict:
					if entDict[entIndex]['classname'] == 'hostage_entity' or entDict[entIndex]['classname'] == 'func_bomb_target': es.server.queuecmd('es_xremove %s' % entIndex)
				self.overtime = 0
				self.timeElapsed = 0
				self.pointsT = 0
				self.pointsCT = 0
				self.roundStarted = True
				pointA.setTeam(1)
				pointB.setTeam(1)
				self.setScreenOverlay()
				delay.add('gametime',1)
			else:
				es.server.queuecmd('mp_ignore_round_win_conditions 0')

	''' Check if players are touching any checkpoints '''
	def tick(self):
		if pointA.loc != None and pointB.loc != None:
			if self.enabled:
				if config.drawBeams:
					pointA.drawBeam()
					pointB.drawBeam()
				del pointA.hitboxPlayerlist[:]
				del pointB.hitboxPlayerlist[:]
				for player in playerlib.getPlayerList('#alive'):
					userid = player.userid
					loc = es.getplayerlocation(userid)
					pointA.checkHitboxStart(userid,loc)
					pointB.checkHitboxStart(userid,loc)
				pointA.checkHitboxEnd()
				pointB.checkHitboxEnd()

	''' Reduce the game time by 1 second and display it in the HUD '''
	def reduceTime(self):
		if self.enabled:
			if self.overtime == 0:
				timeleft = config.timeLimit*60-self.timeElapsed
				timeleftString = '%02d:%02d' % divmod(timeleft,60)
				if timeleft in DICT_COUNTDOWN and self.dominatingTeam == 0: announcer.playSound(DICT_COUNTDOWN[timeleft])
				if timeleft <= 0:
					if self.pointsT == self.pointsCT:
						announcer.playSound('overtime')
						self.centermsg('OVERTIME')
						self.overtime = 1
					elif self.pointsT > self.pointsCT:
						self.forceWin(2)
					else:
						self.forceWin(3)
				elif timeleft < 11:
					if self.dominatingTeam == 0: self.centermsg('%s..' % timeleft)
				self.timeElapsed += 1
			else:
				timeleftString = 'OVERTIME'
			if config.dmEnabled: addonRespawn.respawn.reduceTime()
			if config.showHud:
				for userid in es.getUseridList():
					text = ''
					if config.dmEnabled:
						if userid in addonRespawn.respawn.delays:
							text = 'Respawning in: %s' % addonRespawn.respawn.delays[userid] if addonRespawn.respawn.mode == 1 else 'Respawning in: %s' % addonRespawn.respawn.waveTimeCur
					if config.mode == 1:
						text = '%s \n \nRed Score: %s\nBlue Score: %s\n \nPoint limit: %s' % (text,self.pointsT,self.pointsCT,config.pointLimit)
					elif config.mode == 2:
						text = '%s \n \nRed Score: %s\nBlue Score: %s\n \nTime left: %s' % (text,self.pointsT,self.pointsCT,timeleftString)
					elif config.mode == 3:
						text = '%s \n \nRed Score: %s\nBlue Score: %s\n \nPoint limit: %s\nTime left: %s' % (text,self.pointsT,self.pointsCT,config.pointLimit,timeleftString)
					if text != '': usermsg.hudhint(userid,text)
			delay.add('gametime',1)

	''' Execute the win function for team 2 (T) or 3 (CT) and end the map '''
	def forceWin(self,team):
		if self.enabled:
			self.enabled = 0
			delay.removeAllExcept('tick')
			if team == 2:
				announcer.playSound('fanfare')
				announcer.playSound('winner_red')
				for userid in es.getUseridList(): es.cexec(userid,'r_screenoverlay overlays/ruuubi/dd/winbanner_t')
			elif team == 3:
				announcer.playSound('fanfare')
				announcer.playSound('winner_blue')
				for userid in es.getUseridList(): es.cexec(userid,'r_screenoverlay overlays/ruuubi/dd/winbanner_ct')
			if self.pointsT == 0 or self.pointsCT == 0: delay.add('flawlessVictory',3)
			tempUserid = es.getuserid()
			if tempUserid:
				index = es.createentity('game_end')
				es.server.queuecmd('es_xspawnentity %s' % index)
				es.server.queuecmd('es_xfire %s game_end EndGame' % tempUserid)
			else:
				es.server.queuecmd('changelevel_next')

	''' Update the screen and display the status of checkpoint A+B'''
	def setScreenOverlay(self):
		overlay = '%s|%s' % (pointA.team,pointB.team) if game.enabled else '-1|-1'
		overlay = self.screenoverlays[overlay] if overlay in self.screenoverlays else '0'
		for player in es.getUseridList(): es.cexec(player,'r_screenoverlay %s' % overlay)

	''' Trigger after a checkpoint changed its owner => Outcome: Simple change of owner / team dominates / averted '''
	def changedOwner(self):
		self.setScreenOverlay()
		if pointA.team == pointB.team:
			self.dominatingTeam = pointA.team
			self.dominatingTeamText = 'Red' if self.dominatingTeam == 2 else 'Blue'
			game.centermsg('%s team dominating!' % self.dominatingTeamText)
			announcer.playSound('dominating_%s' % self.dominatingTeamText.lower())
			teamFilter = '#t' if self.dominatingTeam == 2 else '#ct'
			for userid in playerlib.getPlayerList(teamFilter):
				if config.scoreDominate: es.server.queuecmd('score add %i %i'% (userid,config.scoreDominate))
				if config.cashDominate: playerlib.getPlayer(userid).cash += config.cashDominate
			delay.add('dominatingCountdown',2)
		else:
			if self.dominatingTeam != 0: self.averted()

	''' Avert a dominating team '''
	def averted(self):
		if self.dominatingTime > 1:
			game.centermsg('Averted!')
			announcer.playSound('averted')
		else:
			game.centermsg('Last second save!')
			announcer.playSound('last_second_save')
		delay.remove('dominatingCountdown')
		self.reset()

	''' Add a point to the dominating team and start a new round / end the game '''
	def teamScores(self):
		winningTeam = 0
		lead = 0
		if self.dominatingTeam == 2:
			self.pointsT += 1
			lead = self.pointsT - self.pointsCT
			teamFilter = '#t'
			if ((config.mode == 1 or config.mode == 3) and self.pointsT >= config.pointLimit) or self.overtime: winningTeam = 2
		else:
			self.pointsCT += 1
			lead = self.pointsCT - self.pointsT
			teamFilter = '#ct'
			if ((config.mode == 1 or config.mode == 3) and self.pointsCT >= config.pointLimit) or self.overtime: winningTeam = 3
		for userid in playerlib.getPlayerList(teamFilter):
			if config.scoreScore: es.server.queuecmd('score add %i %i'% (userid,config.scoreScore))
			if config.cashScore: playerlib.getPlayer(userid).cash += config.cashScore
		pointA.setTeam(0)
		pointB.setTeam(0)
		self.setScreenOverlay()
		if winningTeam == 0:
			if lead >= 2:
				game.centermsg('%s team increases their lead!' % game.dominatingTeamText)
				announcer.playSound('lead_increase_%s' % self.dominatingTeamText.lower())
			elif lead == 1:
				game.centermsg('%s team takes the lead!' % game.dominatingTeamText)
				announcer.playSound('lead_take_%s' % self.dominatingTeamText.lower())
			else:
				game.centermsg('%s team scores!' % game.dominatingTeamText)
				announcer.playSound('score_%s' % self.dominatingTeamText.lower())
			announcer.playSound('fanfare')
			delay.add('newRound',3)
		else:
			self.forceWin(winningTeam)

	''' Start a new round after a team scored '''
	def newRound(self):
		self.reset()
		pointA.setTeam(1)
		pointB.setTeam(1)
		self.setScreenOverlay()

	''' Send a private message to a player '''
	def whisper(self,userid,msg):
		msg = '#orange[#blueDouble Domination#orange] %s' % str(msg)
		msg = msg.replace('#blue','\x0703B0FF').replace('#red','\x07FF3D3D').replace('#orange','\x07FF8D00')
		es.tell(userid,'#multi',msg)

	''' Check the current version '''
	def checkVersion(self):
		try:
			file = urllib2.urlopen('http://www.nextgenerationgaming.de/addons/double_domination/version.txt')
			version = file.readline()
			file.close()
			return version
		except:
			return None

	''' Display a message in the center of the screen '''
	def centermsg(self,msg):
		for userid in es.getUseridList():
			if players[userid].text: usermsg.centermsg(userid,msg)

	''' Play a hint sound (Deny/Access) '''
	def playHintSound(self,userid,number):
		if number == 0: es.playsound(userid,'buttons/button11.wav',1.0)
		elif number == 1: es.playsound(userid,'UI/buttonclick.wav',1.0)

	''' Trigger the admin popup '''
	def triggerPopupAdmin(self,userid,args):
		if es.getplayersteamid(userid) in config.admins:
			self.playHintSound(userid,1)
			self.popupOpen(userid,'ddAdmin')
		else:
			self.playHintSound(userid,0)
			self.whisper(userid,'#redYou have no access!')

	''' Save the database if saving is enabled'''
	def saveDatabase(self):
		if config.announcerSavetime > 0: cPickle.dump(database,open(PATH_DB,"wb"))

	''' Open a popup '''
	def popupOpen(self,userid,popupname):
		if popupname == 'ddAdmin':
			temp = popuplib.easymenu('ddAdmin',None,self.popupSelected)
			temp.settitle('Double Domination %s' % info.version)
			temp.setdescription('by Ruuubi\n-----------------------------\nMap: %s\nPoint A: %s\nPoint B: %s' % (es.ServerVar('eventscripts_currentmap'),pointA.loc,pointB.loc))
			temp.addoption('setA','[X] Set Point A') if pointA.loc != None else temp.addoption('setA','[  ] Set Point A')
			temp.addoption('setB','[X] Set Point B') if pointB.loc != None else temp.addoption('setB','[  ] Set Point B')
			temp.addoption('save','Save & Restart Map')
			temp.addoption('version','Check Version')
			temp.send(userid)
			temp.delete()
		elif popupname == 'ddPlayer':
			temp = popuplib.easymenu('ddPlayer',None,self.popupSelected)
			temp.settitle('Double Domination %s' % info.version)
			temp.setdescription('by Ruuubi')
			temp.addoption('text','[X] Show Text') if players[userid].text else temp.addoption('text','[  ] Show Text')
			temp.addoption('announcerOFF','[X] Announcer: OFF') if players[userid].announcer == None else temp.addoption('announcerOFF','[  ] Announcer: OFF')
			for index in announcer.announcerDict:
				temp.addoption(index,'[X] Announcer: %s' % announcer.announcerDict[index]) if players[userid].announcer == index else temp.addoption(index,'[  ] Announcer: %s' % announcer.announcerDict[index])
			temp.send(userid)
			temp.delete()

	''' Popup handler '''
	def popupSelected(self,userid,choice,popupname):
		if popupname == 'ddAdmin':
			if choice != 'save': self.playHintSound(userid,1)
			if choice == 'setA':
				self.enabled = False
				x,y,z = es.getplayerlocation(userid)
				pointA.setLoc(list((int(x),int(y),int(z)-10)))
				pointA.setTeam(1)
			elif choice == 'setB':
				self.enabled = False
				x,y,z = es.getplayerlocation(userid)
				pointB.setLoc(list((int(x),int(y),int(z)-10)))
				pointB.setTeam(1)
			elif choice == 'save':
				if pointA.loc != None and pointB.loc != None:
					self.playHintSound(userid,1)
					curMap = str(es.ServerVar('eventscripts_currentmap'))
					iniFile = ConfigParser.RawConfigParser()
					iniFile.read(PATH_ADDON+'/database/mapdata.ini')
					if curMap not in iniFile.sections(): iniFile.add_section(curMap)
					iniFile.set(curMap,'checkpoint_a',','.join(map(str,pointA.loc)))
					iniFile.set(curMap,'checkpoint_b',','.join(map(str,pointB.loc)))
					ini = open(PATH_ADDON+'/database/mapdata.ini','wb')
					iniFile.write(ini)
					self.whisper(userid,'Saving complete!')
					es.server.queuecmd('mp_restartgame 1')
				else:
					self.playHintSound(userid,0)
					self.whisper(userid,'#redCan not save data!')
					self.whisper(userid,'#redAt least one location is not set!')
			elif choice == 'version':
				self.whisper(userid,'Checking version...')
				latestVersion = self.checkVersion()
				if latestVersion == info.version:
					self.whisper(userid,'#greenYou have the latest version! (#blue%s#green)' % info.version)
				elif latestVersion == None:
					self.whisper(userid,"#redVersion check failed!")
					self.whisper(userid,"#redEither the website is down or your firewall is blocking it!")
				else:
					self.whisper(userid,'#redYour version differs from the server!')
					self.whisper(userid,'Your version: #blue%s #orange| Latest: #blue%s' % (info.version,latestVersion))
			self.popupOpen(userid,'ddAdmin')
		elif popupname == 'ddPlayer':
			if choice == 'text':
				self.playHintSound(userid,1)
				players[userid].text = True if not players[userid].text else False
				database[players[userid].steamid]['text'] = players[userid].text
			elif choice == 'announcerOFF':
				self.playHintSound(userid,1)
				players[userid].announcer = None
				database[players[userid].steamid]['announcer'] = None
			else:
				players[userid].announcer = choice
				database[players[userid].steamid]['announcer'] = choice
				announcer.playPreview(userid)
			self.popupOpen(userid,'ddPlayer')

''' Handles announcer sounds '''
class AnnouncerManager(object):
	''' Initialize announcer dictionary '''
	def __init__(self):
		self.announcerDict = {}
		self.sounds = {}
		self.loadSoundData()

	''' Load announcer.ini and add announcer sounds to the dictionary '''
	def loadSoundData(self):
		iniFile = ConfigParser.ConfigParser()
		iniFile.optionxform = str
		iniFile.read(PATH_ADDON+'/configs/announcer.ini')
		for announcer in iniFile.options('Announcer List'): self.announcerDict[iniFile.get('Announcer List',announcer)] = announcer
		for section in iniFile.sections():
			if section != 'Announcer List' and section in self.announcerDict:
				self.sounds[section] = {}
				for option in iniFile.options(section):
					self.sounds[section][option] = iniFile.get(section,option)
					downloadList.append('sound/'+self.sounds[section][option])

	''' Play a soundfile to every user and use their specific announcer '''
	def playSound(self,soundname):
		if config.announcerEnabled == 0: return
		for userid in es.getUseridList():
			announcer = players[userid].announcer
			if announcer != None and announcer in self.sounds:
				if soundname in self.sounds[announcer]: es.playsound(userid,self.sounds[announcer][soundname],1.0)

	''' Play a sound preview to a player to demonstrate the announcer '''
	def playPreview(self,userid):
		announcer = players[userid].announcer
		soundname = 'dominating_red'
		if announcer in self.sounds:
			if soundname in self.sounds[announcer]: es.playsound(userid,self.sounds[announcer][soundname],1.0)

#Lists:
downloadList = ['materials/overlays/ruuubi/dd/info_blue_blue.vmt',
				'materials/overlays/ruuubi/dd/info_blue_blue.vtf',
				'materials/overlays/ruuubi/dd/info_blue_red.vmt',
				'materials/overlays/ruuubi/dd/info_blue_red.vtf',
				'materials/overlays/ruuubi/dd/info_blue_white.vmt',
				'materials/overlays/ruuubi/dd/info_blue_white.vtf',
				'materials/overlays/ruuubi/dd/info_red_blue.vmt',
				'materials/overlays/ruuubi/dd/info_red_blue.vtf',
				'materials/overlays/ruuubi/dd/info_red_red.vmt',
				'materials/overlays/ruuubi/dd/info_red_red.vtf',
				'materials/overlays/ruuubi/dd/info_red_white.vmt',
				'materials/overlays/ruuubi/dd/info_red_white.vtf',
				'materials/overlays/ruuubi/dd/info_white_blue.vmt',
				'materials/overlays/ruuubi/dd/info_white_blue.vtf',
				'materials/overlays/ruuubi/dd/info_white_red.vmt',
				'materials/overlays/ruuubi/dd/info_white_red.vtf',
				'materials/overlays/ruuubi/dd/info_white_white.vmt',
				'materials/overlays/ruuubi/dd/info_white_white.vtf',
				'materials/overlays/ruuubi/dd/winbanner_ct.vmt',
				'materials/overlays/ruuubi/dd/winbanner_ct.vtf',
				'materials/overlays/ruuubi/dd/winbanner_t.vmt',
				'materials/overlays/ruuubi/dd/winbanner_t.vtf',
				'materials/models/ruuubi/dd/white.vmt',
				'materials/models/ruuubi/dd/white.vtf',
				'models/ruuubi/dd/beta_checkpoint_a.dx80.vtx',
				'models/ruuubi/dd/beta_checkpoint_a.dx90.vtx',
				'models/ruuubi/dd/beta_checkpoint_a.mdl',
				'models/ruuubi/dd/beta_checkpoint_a.sw.vtx',
				'models/ruuubi/dd/beta_checkpoint_a.vvd',
				'models/ruuubi/dd/beta_checkpoint_b.dx80.vtx',
				'models/ruuubi/dd/beta_checkpoint_b.dx90.vtx',
				'models/ruuubi/dd/beta_checkpoint_b.mdl',
				'models/ruuubi/dd/beta_checkpoint_b.sw.vtx',
				'models/ruuubi/dd/beta_checkpoint_b.vvd']

#Constants:
PATH_ADDON = es.getAddonPath(info.basename)
PATH_DB = PATH_ADDON+'/database/playersettings.db'
PATH_CHECKPOINT_A = 'models/ruuubi/dd/beta_checkpoint_a.mdl'
PATH_CHECKPOINT_B = 'models/ruuubi/dd/beta_checkpoint_b.mdl'
DICT_COUNTDOWN = {300:'timeleft_5',240:'timeleft_4',180:'timeleft_3',120:'timeleft_2',60:'timeleft_1',30:'countdown_30',20:'countdown_20',10:'countdown_10',9:'countdown_09',8:'countdown_08',7:'countdown_07',6:'countdown_06',5:'countdown_05',4:'countdown_04',3:'countdown_03',2:'countdown_02',1:'countdown_01'}

#Globals:
config = ConfigManager()
pointA = Checkpoint('A')
pointB = Checkpoint('B')
delay = DelayManager()
players = PlayerManager()
game = GameManager()
announcer = AnnouncerManager()

#Database:
database = {}
try: database = cPickle.load(open(PATH_DB,"rb"))
except: pass

#Subaddons:
if config.dmEnabled:
	es.load('double_domination/subaddons/respawn')
	addonRespawn = es.import_addon('double_domination/subaddons/respawn')

if config.spEnabled: es.load('double_domination/subaddons/protection')

#Game Events:
def load():
	es.set('double_domination',info.version)
	es.makepublic('double_domination')
	if len(database) > 0:
		curTime = int(time.time())
		for player in dict(database):
			if 'lastcon' in database[player]:
				if (curTime-database[player]['lastcon'])/86400 >= config.announcerSavetime: del database[player]
		cPickle.dump(database,open(PATH_DB,"wb"))
	cmdlib.registerSayCommand('!ddadmin',game.triggerPopupAdmin,'')
	delay.add('tick',0.1)
	game.setDownloads()
	game.startRound()
	es.msg('#multi','\x07FF8D00[\x0703B0FFDouble Domination\x07FF8D00] \x0703B0FF%s \x07FF8D00successfully loaded!' % info.version)

def unload():
	delay.removeAll()
	game.saveDatabase()
	cmdlib.unregisterSayCommand('!ddadmin')
	es.server.queuecmd('mp_ignore_round_win_conditions 0')
	es.unload('double_domination/subaddons/respawn')
	es.unload('double_domination/subaddons/protection')
	es.msg('#multi','\x07FF8D00[\x0703B0FFDouble Domination\x07FF8D00] \x0703B0FF%s \x07FF8D00successfully unloaded!' % info.version)

def server_shutdown(ev):
	game.saveDatabase()

def es_map_start(ev):
	game.saveDatabase()

def player_spawn(ev):
	userid = int(ev['userid'])
	if int(ev['es_userteam']) > 1:
		if config.cashSpawn: playerlib.getPlayer(userid).cash += config.cashSpawn
		if game.enabled and not game.roundStarted: game.startRound()

def round_start(ev):
	delay.removeAllExcept('tick')
	game.startRound()

def es_map_start(ev):
	game.setDownloads()
	delay.removeAllExcept('tick')

def player_say(ev):
	userid = int(ev['userid'])
	text = ev['text']
	steamid = ev['es_steamid']
	if text.lower() in config.announcerCommands: game.popupOpen(userid,'ddPlayer')

def player_disconnect(ev):
	del players[int(ev['userid'])]